import clips
from expr import BaseExpr, normalize_expr, SymbolExpr
from typedefs import Integer, Number, String, Symbol, FactIndexType, Boolean, Multifield
from utils import LispExpr

class ClipsFuncExpr(BaseExpr):
    def __init__(self, func_name, return_type, *args):
        self.func_name = func_name
        self.return_type = return_type
        self.args = [normalize_expr(arg) for arg in args]
        all_fields = reduce(set.union, (arg.all_fields for arg in self.args), set())
        super(ClipsFuncExpr, self).__init__(all_fields=all_fields)
        
    def get_type(self):
        return self.return_type
        
    def to_lisp(self):
        if not self.args:
            return '({})'.format(self.func_name)
        else:
            return LispExpr(self.func_name, *(arg.to_lisp() for arg in self.args))
    
    def replace_fields(self, field_map):
        args = [arg.replace_fields(field_map) for arg in self.args]
        return self._duplicate_with_args(*args)

    def evaluate(self):
        return clips.Eval(str(self.to_lisp())).encode()
    
    def _duplicate_with_args(self, *args):
        return ClipsFuncExpr(self.func_name, self.return_type, *args)
    
    def __str__(self):
        return '{}({})'.format(self.func_name, ','.join(map(str, self.args)))
    
class ConcatenateExpr(ClipsFuncExpr):
    def __init__(self, *args):
        return super(ConcatenateExpr, self).__init__('', Multifield, *args)
    
    def _duplicate_with_args(self, *args):
        return ConcatenateExpr(*args)
        
    def to_lisp(self):
        return ' '.join(str(arg.to_lisp()) for arg in self.args)
     
class PythonFuncExpr(ClipsFuncExpr):
    def __init__(self, python_func_name, return_type, *args):
        args = (SymbolExpr(python_func_name), ) + args
        super(PythonFuncExpr, self).__init__('python-call', return_type, *args)
        
    def to_lisp(self):
        if self.return_type is Boolean:
            return (self != 0).to_lisp()
        else:
            return super(PythonFuncExpr, self).to_lisp()
        
def clips_func(funcname, return_type):
    def func(*args):
        return ClipsFuncExpr(funcname.replace('_', '-'), return_type, *args)
    func.__name__ = funcname
    return func

# TODO: Add more functions
min_ = clips_func('min', Number)
max_ = clips_func('max', Number)
fact_index = clips_func('fact-index', FactIndexType)
gensym = clips_func('gensym', Symbol)
create_multifield = clips_func('create$', Multifield)
insert_multifield = clips_func('insert$', Multifield)
length_multifield = clips_func('length$', Integer)

def concatenate(*args): return ConcatenateExpr(*args)



"""
This module defines the possible actions that can be used in rule activations. 
"""

from expr import normalize_expr, BaseExpr
from utils import LispExpr, RuleEngineError

class BaseAction(BaseExpr):
    """
    Base class for rule actions
    """
    def prepare_rule(self, rule):
        """ Hook for manipulating any rule that contains the current action """
        pass
    
    def get_type(self):
        # Actions can't be used as expressions. 
        return None

class AssertOrUpdate(BaseAction):
    """
    Base class for actions that deal with data values (assert / update)
    """
    def __init__(self, **data):
        super(AssertOrUpdate, self).__init__()
        self.data = {key : normalize_expr(value) for key, value in data.iteritems()}

    def get_all_fields(self):
        return reduce(set.union, (v.all_fields for v in self.data.itervalues()), set())
                
    def _replace_fields(self, field_map):
        return {key: value.replace_fields(field_map) for key, value in self.data.iteritems()}
    
    def _field_values_to_lisp(self):
        return [LispExpr(key, value.to_lisp()) for key, value in sorted(self.data.iteritems())]

class Assert(AssertOrUpdate):
    """ 
    Defines the most common action: assert (i.e., create) a new fact with given field values.
    If a target template is not defined for the rule, we automatically create one,
    deducing its fields from the Assert parameters.
    """

    def __init__(self, **data):
        super(Assert, self).__init__(**data)
        self.rule = None

    def replace_fields(self, field_map):
        res = Assert(**self._replace_fields(field_map))
        res.rule = self.rule
        return res
    
    def prepare_rule(self, rule):
        if self.rule is not None:
            raise RuleEngineError('"Assert" instance may only be used in a single rule.')
        self.rule = rule

        # Deduce fields for rule target
        if rule.target_fields is None:
            implied_fields = {key:value.get_type() for key, value in self.data.iteritems()}
            rule.set_target_fields(**implied_fields)
        # Fields of input facts cannot be used directly in the Assert() clause.
        # They must be bound to variables.
        for key, value in self.data.iteritems():
            for field in value.all_fields:
                rule.add_variable(field)
        
    def to_lisp(self):
        field_values = self._field_values_to_lisp()
        return LispExpr('assert', LispExpr(self.rule.get_target_name(), *field_values))

    def __str__(self):
        return 'Assert({})'.format(', '.join('{}={}'.format(k, v) for k, v in sorted(self.data.iteritems())))

class Update(AssertOrUpdate):
    """ 
    Defines an action that updates one of the rule's input facts with given field values
    """

    def __init__(self, _fact, **data):
        super(Update, self).__init__(**data)
        self.fact = _fact

    def replace_fields(self, field_map):
        return Update(field_map[self.fact._to_expr()], **self._replace_fields(field_map))
    
    def to_lisp(self):
        field_values = self._field_values_to_lisp()
        return LispExpr('modify', self.fact, *field_values)
    
class Delete(BaseAction):
    """ Defines an action that deletes (retracts) one of the rule's input facts """
     
    def __init__(self, _fact):
        self.fact = _fact
        
    def get_all_fields(self):
        return set()

    def replace_fields(self, field_map):
        return Delete(field_map[self.fact._to_expr()])

    def to_lisp(self):
        return LispExpr('retract', self.fact)



from bunch import Bunch
from functools import wraps

from aggregations import grouping_aggregator
from expr import ConvertibleToExpr
from fact import FactReference
from rule import Rule

class RuleDef(FactReference, ConvertibleToExpr):
    _all_instances = []
    def __init__(self, *modifiers):
        self._rule = Rule()
        for modifier in modifiers:
            modifier(self._rule)
        self._fields = dict(self._rule.get_target_fields() or {})
        self.__dict__.update(self._fields)
        self._all_instances.append(self)
        
    def _set_name(self, name):
        self._rule.set_name(name)
        
    def _build(self, engine):
        self._rule.build(engine)
        
    def _python_action(self, function):
        self._rule.add_python_action(function)
        
    def _aggregator(self, aggregator_cls):
        self._rule.add_aggregator_cls(aggregator_cls)
        
    def _groupby(self, *keys):
        return lambda func : self._rule.add_aggregator_cls(grouping_aggregator(keys, func))
    
    def __getitem__(self, key):
        return self._rule.target[key]
    
    def _to_expr(self):
        return self._rule.target._to_expr()
    
    @property
    def _name(self):
        return self._rule.get_target_name()
    
    def _update_python_param(self, params, value):
        return self._rule.target._update_python_param(params, value)

    def _from_python_params(self, params):
        return self._rule.target._from_python_params(params)

def rule_modifier(method):
    def modifier(*args, **kwargs):
        return wraps(method)(lambda rule : method(rule, *args, **kwargs))
    return modifier

match = rule_modifier(Rule.add_variable)
action = rule_modifier(Rule.add_action)
condition = rule_modifier(Rule.add_condition)
not_exists = rule_modifier(Rule.set_negative_premise)
fields = rule_modifier(Rule.set_target_fields)
salience = rule_modifier(Rule.set_salience)
name = rule_modifier(Rule.set_name)
description = rule_modifier(Rule.set_description)
groupby = rule_modifier(Rule.set_groupby)

@rule_modifier
def target(rule, target):
    if isinstance(target, RuleDef):
        target = target._rule.target
    rule.set_target(target)

@rule_modifier
def foreach(rule, *containers):
    for container in containers:
        rule.set_premise(container)

# Parameters passed from outside world to rule defintions
params = Bunch()

def delayed_set(field, from_value, to_value):
    from actions import Update
    RuleDef(
        condition(field == from_value),
        salience(-1000),
        action(Update(field.container, **{field.name: to_value}))
    )._set_name('AutoSet_{}'.format(field.name)),



from actions import BaseAction, Assert, Update, Delete
from aggregations import Aggregator
from builtin_aggregations import *
from class_ import Class, InstanceField
from clips_func import *
from conditional_expr import if_then_else
from expr import (IntegerField, NumberField, StringField, UnicodeField, SymbolField,
                  BooleanField, DateTimeField, MultiField, TRUE, FALSE, UNKNOWN_BOOL)
from fact import Fact
from func import (RuleFunc, TypedRuleFunc, BooleanRuleFunc, IntegerRuleFunc,
                  NumberRuleFunc, StringRuleFunc)
from operators import and_, or_, not_
from ruledef import (RuleDef, match, action, condition, not_exists, fields,
                     salience, name, description, target, foreach, groupby)
from typedefs import (RuleEngineType, Boolean, DateTime, Integer, Number, 
                      String, Unicode, Multifield) 

__all__ = sorted(x for x in locals() if not x.startswith('_'))



from actions import Assert, Delete, Update
from clips_func import fact_index, gensym, min_, max_, create_multifield, concatenate
from conditional_expr import if_then_else
from expr import normalize_expr, BaseExpr
from rule import Rule

UNIQUE_INDEX_FIELD = '_uniqueIndex'

class MapReduceFunc(BaseExpr):
    def _map(self, *x): return x
    def _reduce(self, x, y): raise NotImplementedError
    
    def __init__(self, *params):
        self.params = map(normalize_expr, params)
        all_fields = set().union(*(param.all_fields for param in self.params))
        super(MapReduceFunc, self).__init__(all_fields=all_fields)
        map_values = self._map(*self.params)
        if not isinstance(map_values, (tuple, list)):
            map_values = [map_values]
        self.map_values = map(normalize_expr, map_values)
        
    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, 
                               ', '.join(str(param) for param in self.params))
        
    def __repr__(self):
        return '<{}>'.format(self)
        
    def get_type(self):
        return self.params[0].get_type()

class AssertAggregate(Assert):
    def __init__(self, **kwargs):
        self.map_reduce_values = {}
        for key, value in kwargs.items():
            if isinstance(value, MapReduceFunc):
                del kwargs[key]
                subkeys = []
                for n, map_value in enumerate(value.map_values):
                    subkey = key if n == 0 else '{}_{}'.format(key, n+1)
                    kwargs[subkey] = map_value
                    subkeys.append(subkey)
                self.map_reduce_values[key] = (value, subkeys)
        super(AssertAggregate, self).__init__(**kwargs)
        
    def prepare_rule(self, rule):
        self.data.update({key : normalize_expr(value) for key, value in rule.groupby.iteritems()})
        self.data[UNIQUE_INDEX_FIELD] = gensym()  
        super(AssertAggregate, self).prepare_rule(rule)
        self._create_reduce_rule(rule)
        
    def _create_reduce_rule(self, map_rule):
        r1 = map_rule.target[0]
        r2 = map_rule.target[1]
        self.reduce_rule = Rule()
        self.reduce_rule.set_name('_Reduce')
        for field_name in map_rule.groupby:
            self.reduce_rule.add_variable(getattr(r1, field_name), getattr(r2, field_name))
        self.reduce_rule.add_condition(fact_index(r1) != fact_index(r2))
        
        reduce_kwargs = {}
        for field_name, (value, subfields) in self.map_reduce_values.iteritems():
            v1 = [getattr(r1, field_name) for field_name in subfields]
            v2 = [getattr(r2, field_name) for field_name in subfields]
            if len(subfields) == 1: 
                v1 = v1[0]
                v2 = v2[0]
            reduced = value._reduce(v1, v2)
            if len(subfields) == 1:
                reduced = [reduced]
            reduce_kwargs.update(dict(zip(subfields, reduced)))
        self.reduce_rule.add_action(Delete(r2))
        self.reduce_rule.add_action(Update(r1, **reduce_kwargs))
        self.reduce_rule.set_salience(map_rule.salience)
        map_rule.add_secondary_rule(self.reduce_rule)
        
class Sum(MapReduceFunc): 
    def _reduce(self, x, y): return x + y

class Max(MapReduceFunc):
    def _reduce(self, x, y): return max_(x, y)
    
class Min(MapReduceFunc):
    def _reduce(self, x, y): return min_(x, y)
    
class Concatenate(MapReduceFunc):
    def _map(self, x):
        return create_multifield(x)
     
    def _reduce(self, x, y):
        return concatenate(x, y)

class BaseArgMinMax(MapReduceFunc):
    def __init__(self, x, key):
        return super(BaseArgMinMax, self).__init__(x, key)
    
    def _reduce(self, x, y):
        condition = self._condition(x[1], y[1]) 
        return (if_then_else(condition, x[0], y[0]), if_then_else(condition, x[1], y[1]))

class ArgMin(BaseArgMinMax):
    def _condition(self, x, y): return x <= y

class ArgMax(BaseArgMinMax):
    def _condition(self, x, y): return x >= y
    
class Count(Sum):
    def _map(self): return 1
    
__all__ = 'AssertAggregate Sum Max Min ArgMax ArgMin Count Concatenate'.split()



"""
This module provides the mechanism for wrapping user-defined Python functions
so that they can be called inside rules (in conditions, assertions etc.)
"""

import clips
import functools
import sys
from clips_func import PythonFuncExpr
from typedefs import Boolean, String, Integer, Number
from utils import UniqueIdCounter, logger

class RuleFunc(object):
    """
    Wrap a Python function to be usable inside the rule engine
    """
    _last_error = None
    _counters = {}
    _engine_by_environment = {}
    _cur_engine = None
    NUM_CALLS_PER_PRINT = 1000
    
    def __init__(self, func, return_type=String, func_name=None):
        self.func = func
        self.type = return_type
        self.func_name = func_name or func.__name__
        self._wrapper_by_engine = {}
        self._num_calls = 0
        
    def __call__(self, *args, **kwargs):
        """
        Return an expression that indicates an invocation of this function with
        given arguments.
        """
        wrapper = self._wrapper_by_engine.get(self._cur_engine)
        if wrapper is None:
            wrapper = self._create_wrapper(self._cur_engine)
            clips.RegisterPythonFunction(wrapper)
        return PythonFuncExpr(wrapper.__name__, self.type, *args)
    
    def _create_wrapper(self, engine):
        """
        Creates a wrapper Python function that would be run by PyCLIPS.
        It calls the user defined function and is also responsible for handling
        errors and for type conversions.
        """
        def wrapper(*args):
            if self._last_error is not None:
                return
            try:
                args = [(engine._wrap_clips_instance(arg) if isinstance(arg, 
                    (clips.Fact, clips.InstanceName)) else arg) for arg in args]
                res = self.func(*args)
                self._num_calls += 1
                if self._num_calls % self.NUM_CALLS_PER_PRINT == 0:
                    logger.debug('{}: called {} times.'.format(self.func_name, self._num_calls))
                return self.type._to_clips_value(res)
            except:
                # Register error to be raised later by engine
                cls = type(self)
                type(self)._last_error = sys.exc_info()
                # Abort CLIPS execution
                engine.environment.Clear()
        wrapper.__name__ = self._counters.setdefault(self.func_name, UniqueIdCounter(self.func_name+'_')).next()
        return wrapper
    
    # Global methods for internal use

    @classmethod
    def _register_engine(cls, engine):
        cls._engine_by_environment[engine.environment] = engine
        
    @classmethod
    def _set_cur_engine(cls, engine):
        cls._cur_engine = engine
            
    @classmethod    
    def _clear_error(cls):
        cls._last_error = None
        
    @classmethod
    def _check_error(cls):
        if cls._last_error:
            raise cls._last_error[0], cls._last_error[1], cls._last_error[2]

TypedRuleFunc = lambda return_type : functools.partial(RuleFunc, return_type=return_type)

# Shortcuts for defining functions with different return types
BooleanRuleFunc = TypedRuleFunc(Boolean)
IntegerRuleFunc = TypedRuleFunc(Integer)
NumberRuleFunc = TypedRuleFunc(Number)
StringRuleFunc = TypedRuleFunc(String)




import clips
import os

from def_module_loader import DefModuleLoader
from fact import Fact, RULU_INTERNAL_PREFIX
from func import RuleFunc
from utils import RuleEngineError, logger


class RuleEngine(object):
    def __init__(self):
        self.environment = clips.Environment()
        self.logger = logger.getChild(type(self).__name__)
        self.clips_types = {}
        self.preprocess_funcs = []
        self.postprocess_funcs = []
        RuleFunc._register_engine(self)

    def load_module(self, module_name, package=None, auto_salience=False, debug_rules=False):
        """ 
        Load data model and rule definitions from the given Python module 
        """
        DefModuleLoader(self).load(module_name, package, auto_salience=auto_salience,
                                   debug_rules=debug_rules)

    def assert_(self, fact_type, **values):
        """ 
        Assert a new fact 
        """
        if isinstance(fact_type, basestring):
            fact_type = self.clips_types.get(fact_type)
        if not issubclass(fact_type, Fact):
            raise TypeError('{} is not a fact type.'.format(fact_type))
        fact = fact_type(**values)
        self.environment.Assert(fact._clips_obj)
        return fact
        
    def run(self):
        self.run_one_cycle()
        
    def run_one_cycle(self, limit_steps=None):
        """
        Run a single cycle of the rule engine
        
        limit_steps: maximal number of execution steps (or None to run 
                     until completion)
        """
        for preprocess_func in self.preprocess_funcs:
            self.logger.info('Calling: %s', preprocess_func)
            preprocess_func()
        RuleFunc._clear_error()
        if limit_steps:
            self.logger.info('Running rule engine (%d steps)', limit_steps)
        else:
            self.logger.info('Running rule engine')
        self.environment.Run(limit_steps)
        self.logger.info('Rule engine completed')
        RuleFunc._check_error()
        for postprocess_func in self.postprocess_funcs:
            self.logger.info('Calling: %s', postprocess_func)
            postprocess_func()
            
    def get_all_facts(self):
        """
        Return an iterator over all known facts
        """
        return (self._wrap_clips_instance(fact) for fact in self.environment.FactList()
                if fact.Relation != 'initial-fact'
                and not fact.Relation.startswith(RULU_INTERNAL_PREFIX))
            
    def get_facts(self, type_name):
        """
        Return an iterator over all known facts of given type
        """
        return (self._wrap_clips_instance(fact) for fact in self.environment.FactList()
                if fact.Relation == type_name)
    
    def clear(self):
        self.environment.Clear()
        
    def reset(self):
        """
        1. Remove all activated rules from agenda
        2. Remove all facts from the fact-list
        3. Assert the facts from existing deffacts
        """
        self.environment.Reset()
        
    def load(self, filename):
        """
        Load facts and class instances from a text file in CLIPS format
        (as written by the save() method)
        """
        try:
            self.logger.debug('Loading facts from {}'.format(filename))
            self.environment.LoadFacts(filename)
            instance_filename = _get_instance_filename(filename)
            if os.path.exists(instance_filename):
                self.environment.LoadInstances(instance_filename)
        except IOError:
            raise RuleEngineError('Error while loading {}.\n Error log:\n{}'.format(filename, clips.ErrorStream.Read()))
        
    def save(self, filename):
        """
        Save all facts and class instances to a text file in CLIPS format
        """
        self.logger.debug('Saving facts to {}'.format(filename))
        self.environment.SaveFacts(filename)
        
        if len(self.environment.DefinstancesList()) > 1: # There is 1 by default
            instance_filename = _get_instance_filename(filename)
            self.logger.debug('Saving instances to {}'.format(instance_filename))
            self.environment.SaveInstances(instance_filename)
        
    def register_clips_type(self, clips_type):
        self.clips_types[clips_type._name] = clips_type
        
    def get_rule_names(self):
        """
        List the names of all the defined rules
        """
        return [str(t).rsplit('::', 1)[1] for t in self.environment.RuleList()]

    def _wrap_clips_instance(self, instance):
        """
        Take a fact/instance as returned from PyCLIPS, and wrap it in the
        appropriate Fact/Class instance.
        """
        if isinstance(instance, clips.InstanceName):
            instance = self.environment.FindInstance(instance)
            return self.clips_types[str(instance.Class)](_clips_obj=instance)
        else:
            return self.clips_types[str(instance.Relation)](_clips_obj=instance)


def _get_instance_filename(fact_filename):
    tokens = fact_filename.rsplit('.', 1)
    tokens[0] += '.instances'
    return '.'.join(tokens)



import sys
sys.path[:0] = ['/Users/ubarkai/repos/rulu']

import logging
from pkg_resources import resource_filename  # @UnresolvedImport
from unittest import TestCase
from rulu.builtin_aggregations import UNIQUE_INDEX_FIELD
from rulu.engine import RuleEngine
from rulu.slots import HasSlots

import inputs
import expected_outputs
import ruledefs
from ruledefs import nearest_distance_rules

logging.basicConfig(level=logging.DEBUG)


class BaseRuleEngineTest(TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)

    def _engine_test(self, ruledef, input, expected_output=None):  # @ReservedAssignment
        engine = self._make_engine(ruledef)
        engine.load(resource_filename(inputs.__name__, input))
        engine.run_one_cycle()
        actual_facts = self._get_facts(engine)

        if expected_output is not None:
            engine2 = self._make_engine(ruledef)
            engine2.load(resource_filename(inputs.__name__, input))
            engine2.load(resource_filename(expected_outputs.__name__, expected_output))
            expected_facts = self._get_facts(engine2)
            self.assertEqual(expected_facts, actual_facts)
        return engine

    def _make_engine(self, ruledef):
        engine = RuleEngine()
        engine.load_module(ruledef, ruledefs)
        return engine

    def _get_facts(self, engine):
        res = []
        for fact in engine.get_all_facts():
            values = fact._as_dict()
            values.pop('_id', None)
            values.pop(UNIQUE_INDEX_FIELD, None)
            for key, value in values.iteritems():
                if isinstance(value, HasSlots):
                    values[key] = value._as_dict()
            res.append('{} / {}'.format(fact._name, values))
        return sorted(res)



class RuleEngineTests(BaseRuleEngineTest):
    maxDiff = None
    
    def testFamily(self):
        self._engine_test(ruledef='family', input='fathers.txt', expected_output='grandfathers.txt')
        
    def testFamilyPython(self):
        self._engine_test(ruledef='family_python', input='fathers.txt', expected_output='grandfathers.txt')
        
    def testFamilyPythonFunc(self):
        self._engine_test(ruledef='family_func', input='fathers.txt', expected_output='grandfathers.txt')
        
    def testPythonFuncError(self):
        def testWithError():
            self._engine_test(ruledef='family_python_error', input='fathers.txt', expected_output=None)
        self.assertRaises(RuntimeError, testWithError)
        
    def testPythonAggregator(self):
        self._engine_test(ruledef='python_aggregations', input='pairs.txt', expected_output='aggregates.txt')
        
    def testBuiltinAggregator(self):
        self._engine_test(ruledef='builtin_aggregations', input='pairs.txt', expected_output='aggregates_with_count.txt')
        
    def testMultifield(self):
        self._engine_test(ruledef='multifield', input='pairs.txt', expected_output='concatenated.txt')
        
    def testUpdateDelete(self):
        self._engine_test(ruledef='update_delete', input='pairs.txt', expected_output='aggregates.txt')
        
    def testUpdateDeletePython(self):
        self._engine_test(ruledef='update_delete_python', input='pairs.txt', expected_output='aggregates.txt')

    def testMatchConst(self):
        self._engine_test(ruledef='match_const', input='consts.txt', expected_output='consts.txt')
        
    def testClassesWithPythonFunc(self):
        nearest_distance_rules.make_rule = nearest_distance_rules.python_func_distance_rule
        self._engine_test(ruledef='nearest', input='points.txt', expected_output='nearest.txt')
        
    def testClassesWithPythonRule(self):
        nearest_distance_rules.make_rule = nearest_distance_rules.python_distance_rule
        self._engine_test(ruledef='nearest', input='points.txt', expected_output='nearest.txt')
        
    def testClassesPythonFree(self):
        nearest_distance_rules.make_rule = nearest_distance_rules.python_free_distance_rule
        self._engine_test(ruledef='nearest', input='points.txt', expected_output='nearest.txt')



from collections import defaultdict
from fact import IndexedFactReference
from ruledef import RuleDef
from utils import RuleEngineError

def auto_set_salience(ruledefs):
    dependencies = _get_dependencies(ruledefs)
    order = _topological_sort(dependencies)
    for n, rule in enumerate(order):
        rule.set_salience(-(n*2))
    
def _get_dependencies(ruledefs):
    by_target = defaultdict(list)
    for ruledef in ruledefs:
        by_target[ruledef._rule.target].append(ruledef._rule)
        by_target[ruledef._rule.target].extend(ruledef._rule.secondary_rules)

    dependencies = defaultdict(set)
    for ruledef in ruledefs:
        rule = ruledef._rule
        for premise in rule.premises:
            if isinstance(premise, IndexedFactReference):
                premise = premise._template_cls
            if isinstance(premise, RuleDef):
                premise = premise._rule.target
            for other_rule in by_target[premise]:
                if other_rule is not rule:
                    dependencies[other_rule].add(rule)
        dependencies[rule].update(rule.secondary_rules)

    return dependencies

# Adapted from networkx (https://networkx.github.io)
def _topological_sort(dependencies):
    seen = set()
    order = []
    explored = set()

    for v in dependencies.keys():     # process all vertices
        if v in explored:
            continue
        fringe = [v]   # nodes yet to look at
        while fringe:
            w = fringe[-1]  # depth first search
            if w in explored: # already looked down this branch
                fringe.pop()
                continue
            seen.add(w)     # mark as seen
            # Check successors for cycles and for new nodes
            new_nodes = []
            for n in dependencies[w]:
                if n not in explored:
                    if n in seen: #CYCLE !!
                        raise RuleEngineError("Dependencies graph contains a cycle, cannot set salience automatically.")
                    new_nodes.append(n)
            if new_nodes:   # Add new_nodes to fringe
                fringe.extend(new_nodes)
            else:           # No new nodes so w is fully explored
                explored.add(w)
                order.append(w)
                fringe.pop()    # done considering this node
    return list(reversed(order))



"""
Basic wrapper for CLIPS Object-Oriented (COOL) constructs 
"""

import clips
from expr import FieldExpr, BaseExpr
from slots import HasSlots
from typedefs import RuleEngineType
from utils import wrap_clips_errors

class Class(HasSlots, RuleEngineType):
    CLIPS_INSTANCE_TYPE = clips.Instance
    CLIPS_TYPE = 'INSTANCE'
            
    @classmethod
    @wrap_clips_errors
    def _build(cls, engine):
        if cls is not Class:
            super(Class, cls)._build(engine)
            definition = '(is-a USER){}'.format(cls._slots)
            cls._clips_type = engine.environment.BuildClass(cls._name, definition)
        
    def _create_clips_obj(self):
        return self._clips_type.BuildInstance()

    def _copy_clips_obj(self, obj):
        return obj if hasattr(obj, '_Instance__env') else self._environment.Instance(obj)
    
    @classmethod
    def _from_clips_value(cls, x):
        if isinstance(x, clips.InstanceName):
            return cls(_clips_obj=cls._environment.FindInstance(x))
        else: # Instance
            return cls(_clips_obj=x)
    
    @classmethod
    def _isinstance(cls, x):
        return isinstance(x, cls)
    
class InstanceField(FieldExpr):
    def __init__(self, _type):
        super(InstanceField, self).__init__(_type)
        for key, field in _type._fields.iteritems():
            setattr(self, key, InstanceMember(self, key, field))
            
class InstanceMember(BaseExpr):
    def __init__(self, container, key, field):
        super(InstanceMember, self).__init__(all_fields=[container])
        self._container = container
        self._key = key
        self._field = field
        
    def __str__(self):
        return '{}.{}'.format(self._container, self._key)
    
    def get_type(self):
        return self._field._type

    def to_lisp(self):
        return '(send {} get-{})'.format(self._container.to_lisp(), self._key)

    def replace_fields(self, field_map):
        return InstanceMember(self._container.replace_fields(field_map), self._key, self._field)



import sys
from collections import defaultdict
from importlib import import_module
from threading import Lock
from types import ModuleType

from auto_salience import auto_set_salience
from func import RuleFunc
from slots import HasSlots
from ruledef import RuleDef, params
from rule import Rule
from utils import logger

class DefModuleLoader(object):
    def __init__(self, engine):
        self.engine = engine
        self.lock = Lock()
        self.logger = logger.getChild(type(self).__name__)
        
    def load(self, module_name, package=None, auto_salience=False, debug_rules=False, **module_params):
        # Prepare for import
        if isinstance(package, ModuleType):
            package = package.__name__
        if package is not None and '.' not in module_name:
            module_name = '.' + module_name
        self.logger.debug('Loading: %s', module_name)
            
        # Import module, while keeping track of entities that require compilation
        existing_modules = set(sys.modules)
        with self.lock:
            params.update(module_params)
            RuleDef._all_instances = []
            HasSlots._all_subclasses = []
            RuleFunc._set_cur_engine(self.engine)
            module = import_module(module_name, package)
            if module.__name__ in existing_modules:
                reload(module)
            # TODO: Why are there non-named templates??
            templates_to_build = [x for x in HasSlots._all_subclasses if x._name is not None]
            rules_to_build = RuleDef._all_instances

        # Add name to all unnamed instances (according to variable name)
        for name in dir(module):
            entity = getattr(module, name)
            if isinstance(entity, RuleDef) and entity._rule.name is None:
                entity._set_name(name) # TODO: include module name

        if auto_salience:
            auto_set_salience(RuleDef._all_instances)
        if debug_rules:
            self._build_debug_rules()

        # Sort for clarity
        templates_to_build.sort(key=lambda x:x._name)
        if auto_salience:
            rules_to_build.sort(key=lambda x:(-(x._rule.salience or 0), x._rule.name))
                
        # Compile everything, except unnamed templates (which cannot be compiled)
        # Note that RuleDefs also may contain TemplateDefs, which would not be named.
        for entity in templates_to_build + rules_to_build:
            entity._build(self.engine)

    def _build_debug_rules(self):
        by_salience = defaultdict(list)
        for ruledef in RuleDef._all_instances:
            by_salience[ruledef._rule.salience or 0].append(ruledef._rule)
        for salience, rules in sorted(by_salience.iteritems()):
            self._add_debug_rule(salience, rules)
            
    def _add_debug_rule(self, salience, preceding_rules):
        rule_names = ', '.join(rule.name or '->{}'.format(rule.target._name) for rule in preceding_rules)
        def debug_func(*args, **kwargs):
            self.logger.debug('Ran rule(s): {} at salience: {}.'.format(rule_names, salience))
        rule = Rule()
        rule.set_salience(salience-1)
        rule.add_python_action(debug_func)
        rule.build(self.engine)



from utils import RuleEngineError, logger
from progress_bar import LogProgressBar

try:
    import xlwt
except ImportError:
    raise NotImplementedError('xlwt must be installed to support export to Excel')

_logger = logger.getChild(__name__)

class ExportSheetWriter(object):
    def __init__(self, workbook, sheet_name, max_rows_in_sheet):
        self.workbook = workbook
        self.sheet_name = sheet_name[:30]
        self.max_rows = max_rows_in_sheet - 1
        self.nrows = 0
        
    def _init_sheet(self, fact):
        self.sheet = self.workbook.add_sheet(self.sheet_name)
        self.fields = sorted(fact)
        for nfield, field in enumerate(self.fields):
            self.sheet.write(self.nrows, nfield, field)
            
    def write_fact(self, fact):
        if self.nrows == 0:
            self._init_sheet(fact)
        elif self.nrows >= self.max_rows:
            return
        self.nrows += 1
        for nfield, field in enumerate(self.fields):
            self.sheet.write(self.nrows, nfield, fact[field])
            
def read_fact_file(facts_filename):
    with open(facts_filename) as f:
        num_facts = sum(1 for _ in f)
    progress_bar = LogProgressBar(_logger, num_facts, 'Reading facts...', 10000)
    with open(facts_filename) as f:
        for line in f:
            template_name, values = line.strip()[1:-1].split(None, 1)
            tokens = values.split('(')[1:]
            fact = {}
            for token in tokens:
                token = token.strip()
                assert token.endswith(')'), token
                key, value = token[:-1].split(None, 1)
                if value.startswith('"') and value.endswith('"'):
                    value = value[1:-1].replace('\\"', '"')
                fact[key] = value
            yield template_name, fact
            progress_bar.increment()    

def export_facts_to_xls(facts_filename, out_filename, max_rows_in_sheet=10000):
    workbook = xlwt.Workbook()
    writers = {}
    _logger.info('Exporting facts to {}.'.format(out_filename))
    for template_name, fact in read_fact_file(facts_filename):
        writer = writers.get(template_name)
        if writer is None:
            writer = ExportSheetWriter(workbook, template_name, max_rows_in_sheet=max_rows_in_sheet)
            writers[template_name] = writer
        writer.write_fact(fact)
    if not writers:
        raise RuleEngineError('No facts found, cannot export to Excel')
    workbook.save(out_filename)



"""
This module defines the possible types in the rule engine, and provides 
conversions from and to Python types.
"""

from calendar import timegm
from collections import Iterable
from datetime import datetime
import clips

class RuleEngineType(object):
    """
    Base class for a possible type
    """
    PYTHON_TYPE = None
    PYTHON_TYPE_CHECK = None
    CLIPS_TYPE = None
    
    @classmethod
    def _isinstance(cls, x):
        """ Check if argument is of the current type """
        return isinstance(x, cls.PYTHON_TYPE_CHECK)
    
    @classmethod
    def _from_clips_value(cls, x):
        """ Convert from CLIPS value to Python value """
        return cls.PYTHON_TYPE(x)
    
    @classmethod
    def _to_clips_value(cls, x):
        """ Convert from Python value to CLIPS value """
        return x

class Integer(RuleEngineType):
    PYTHON_TYPE = int
    PYTHON_TYPE_CHECK = int
    CLIPS_TYPE = 'INTEGER'
    
class Boolean(Integer):
    PYTHON_TYPE = bool
    PYTHON_TYPE_CHECK = bool
    CLIPS_TYPE = 'SYMBOL'
    TRUE = clips.Symbol('TRUE')
    FALSE = clips.Symbol('FALSE')
    
    @classmethod
    def _to_clips_value(cls, x):
        if isinstance(x, clips.Symbol):
            return x
        elif x:
            return cls.TRUE
        else:
            return cls.FALSE
    
    @classmethod
    def _from_clips_value(cls, x):
        return str(x) == 'TRUE'

class Number(RuleEngineType): 
    PYTHON_TYPE = float
    PYTHON_TYPE_CHECK = (int, float)
    CLIPS_TYPE = 'NUMBER'
    
class String(RuleEngineType):
    PYTHON_TYPE = str
    PYTHON_TYPE_CHECK = basestring
    CLIPS_TYPE = 'STRING'
    
class Unicode(RuleEngineType):
    """
    Unicode string.
    PyCLIPS does not support Unicode values, so we take of the encoding/decoding here 
    """
    ENCODING = 'utf8'
    PYTHON_TYPE = unicode
    PYTHON_TYPE_CHECK = basestring
    CLIPS_TYPE = 'STRING'
    
    @classmethod
    def _to_clips_value(cls, x):
        return x.encode(cls.ENCODING)

    @classmethod
    def _from_clips_value(cls, x):
        return x.decode(cls.ENCODING)
    
class Symbol(String):
    CLIPS_TYPE = 'SYMBOL'
    
class DateTime(RuleEngineType):
    """
    Date-time object. 
    There is no date object in CLIPS, we use an integer representation to
    allow for comparisons and arithmetic.
    """
    PYTHON_TYPE = datetime
    PYTHON_TYPE_CHECK = datetime
    CLIPS_TYPE = 'INTEGER'
    
    @classmethod
    def _to_clips_value(cls, x):
        return timegm(x.utctimetuple())

    @classmethod
    def _from_clips_value(cls, x):
        return datetime.utcfromtimestamp(x)
    
class Multifield(RuleEngineType):
    """
    Multifield type. Maps to Python list.
    """
    PYTHON_TYPE = list
    PYTHON_TYPE_CHECK = Iterable
    CLIPS_TYPE = 'MULTIFIELD'
     
    @classmethod
    def _to_clips_value(cls, x):
        return clips.Multifield(x)
     
class FactIndexType(RuleEngineType): pass

TYPE_MAP = {t.PYTHON_TYPE: t for t in (Integer, Number, String, Boolean, DateTime)}



"""
General utility classes and constructs.
"""

import clips
import functools
import logging
import sys

class RuleEngineError(Exception): pass

def wrap_clips_errors(func):
    """
    Add error details (from CLIPS error stream) to CLIPS exceptions. 
    """
    @functools.wraps(func)
    def new_func(*a, **kw):
        try:
            return func(*a, **kw)
        except clips.ClipsError:
            raise RuleEngineError(clips.ErrorStream.Read()), None, sys.exc_info()[2]
    return new_func

class Printable(object):
    """ 
    Add type name to CLIPS object representation.
    """
    def __str__(self):
        raise NotImplementedError
    
    def __repr__(self):
        return '<{} ({})>'.format(type(self).__name__, self) 

class LispExpr(object):
    """
    Represents a nested LISP-like expression.
    """
    def __init__(self, *values):
        if len(values) == 1:
            self.values = values
        else:
            self.values = []
            for v in values:
                if isinstance(v, tuple): v = LispExpr(*v)
                elif not isinstance(v, LispExpr): v = LispExpr(v)
                self.values.append(v)
        
    def __str__(self):
        """
        Return current expression in LISP format.
        """
        res = ' '.join(str(v) for v in self.values)
        if len(self.values) > 1:
            res = '({})'.format(res)
        return res

    def __repr__(self):
        return '<{} ({})>'.format(type(self), self) 

class UniqueIdCounter(object):
    def __init__(self, prefix):
        self.prefix = prefix
        self.count = 0
        
    def next(self):
        self.count += 1
        return '{}{}'.format(self.prefix, self.count)

logger = logging.getLogger('ruleengine')



from utils import Printable, RuleEngineError
from typedefs import Boolean, DateTime, Integer, Multifield, Number, String, Unicode, Symbol

class BaseExpr(Printable):
    def __init__(self, all_fields=[]):
        self.all_fields = set(all_fields)
        
    def to_lisp(self): raise NotImplementedError
    def evaluate(self): raise NotImplementedError

    def __gt__(self, rhs): return self._boolean_operator(rhs, '>')
    def __ge__(self, rhs): return self._boolean_operator(rhs, '>=')
    def __lt__(self, rhs): return self._boolean_operator(rhs, '<')
    def __le__(self, rhs): return self._boolean_operator(rhs, '<=')
    def __eq__(self, rhs): return self._boolean_operator(rhs, 'eq')
    def __ne__(self, rhs): return self._boolean_operator(rhs, 'neq')
    
    def __add__(self, rhs): return self._binary_operator(rhs, '+')
    def __sub__(self, rhs): return self._binary_operator(rhs, '-')
    def __mul__(self, rhs): return self._binary_operator(rhs, '*')
    def __div__(self, rhs): return self._binary_operator(rhs, '/')
    def __pow__(self, rhs): return self._binary_operator(rhs, '**')
    
    def _binary_operator(self, rhs, op):
        from operators import ArithmeticBinaryOperator
        return ArithmeticBinaryOperator(op, self, normalize_expr(rhs))
    
    def _boolean_operator(self, rhs, op):
        from operators import BooleanBinaryOperator
        return BooleanBinaryOperator(op, self, normalize_expr(rhs))
    
    def replace_fields(self, field_map):
        return field_map.get(self, self)
    
    def get_type(self):
        raise RuleEngineError('Cannot infer type for {}'.format(self))
    
class PrimitiveExpr(BaseExpr):
    def __init__(self, value):
        super(PrimitiveExpr, self).__init__()
        self.original_value = value
        if isinstance(value, bool): 
            self._type = Boolean
            value = Boolean._to_clips_value(value)
        elif isinstance(value, (int, long)): self._type = Integer
        elif isinstance(value, float): self._type = Number
        elif isinstance(value, str): 
            self._type = String
            value = '"{}"'.format(value.replace('"', '\\"'))
        elif isinstance(value, unicode):
            self._type = Unicode
            value = u'"{}"'.format(value.replace(u'"', u'\\"'))
        else:
            raise TypeError(value)
        self.value = value
        
    def __str__(self):
        return repr(self.value)
    
    def to_lisp(self):
        return self.value
    
    def get_type(self):
        return self._type
    
    def evaluate(self):
        return self.original_value
    
class SymbolExpr(BaseExpr):
    def __init__(self, value, type=None):
        super(SymbolExpr, self).__init__()
        self.value = value
        self.type = type
        
    def to_lisp(self):
        return self.value
    
    def __str__(self):
        return self.value
    
    def get_type(self):
        return self.type
    
class FieldExpr(BaseExpr):
    global_order = 0
    
    def __init__(self, _type=None):
        if _type is not None: self._type = _type
        super(FieldExpr, self).__init__(all_fields=[self])
        self.initialized = False
        self.order = FieldExpr.global_order
        FieldExpr.global_order += 1 
        
    def _init(self, container, field_name):
        self.container = container
        self.name = field_name
        self.initialized = True
        
    def __str__(self):
        if self.initialized:
            return '{}.{}'.format(self.container, self.name)
        else:
            return 'uninitialized'

    def get_type(self):
        return self._type
    
class IntegerField(FieldExpr): _type = Integer
class NumberField(FieldExpr): _type = Number
class StringField(FieldExpr): _type = String
class SymbolField(FieldExpr): _type = Symbol
class UnicodeField(FieldExpr): _type = Unicode
class BooleanField(FieldExpr): _type = Boolean
class DateTimeField(FieldExpr): _type = DateTime
class MultiField(FieldExpr): _type = Multifield

class VariableExpr(BaseExpr):
    def __init__(self, var_name, var_type):
        super(VariableExpr, self).__init__()
        self.var_name = var_name
        self.var_type = var_type
        
    def __str__(self):
        return self.var_name
    
    def to_lisp(self):
        return str(self)
    
    def get_type(self):
        return self.var_type
    
class ConvertibleToExpr(object):
    def _to_expr(self):
        raise NotImplementedError
    
def normalize_expr(expr):
    if isinstance(expr, BaseExpr):
        return expr
    elif isinstance(expr, ConvertibleToExpr):
        return expr._to_expr()
    else:
        try:
            return PrimitiveExpr(expr)
        except TypeError:
            raise RuleEngineError('Expected primitive or expression, found {}'.format(expr))

TRUE = SymbolExpr('TRUE', type=Boolean)
FALSE = SymbolExpr('FALSE', type=Boolean)
UNKNOWN_BOOL = SymbolExpr('UNKNOWN', type=Boolean)



from expr import BaseExpr, normalize_expr
from utils import RuleEngineError, LispExpr

class ConditionalExpr(BaseExpr):
    def __init__(self, *args):
        self.args = map(normalize_expr, args)
        types = {arg.get_type() for arg in self.args[1:3]}
        if len(types) != 1:
            raise RuleEngineError("Different return types: {}".format(types))
        self.return_type = types.pop()
        all_fields = reduce(set.union, (arg.all_fields for arg in self.args), set())
        super(ConditionalExpr, self).__init__(all_fields=all_fields)

    def get_type(self):
        return self.return_type
        
    def to_lisp(self):
        args = [arg.to_lisp() for arg in self.args]
        return LispExpr('if', args[0], 'then', args[1], 'else', args[2])
    
    def replace_fields(self, field_map):
        args = [arg.replace_fields(field_map) for arg in self.args]
        return type(self)(*args)
    
    def __str__(self):
        return '({} ? {} : {})'.format(*map(str, self.args))

def if_then_else(condition, then, else_):
    return ConditionalExpr(condition, then, else_)


import logging
from pkg_resources import resource_filename

import inputs
import ruledefs
from rulu.engine import RuleEngine
from rulu.rulu_io import facts_to_df
from unittest import TestCase

logging.basicConfig(level=logging.DEBUG)

class RuleEngineIoTests(TestCase):
    def setUp(self):
        self.engine = RuleEngine()
        self.engine.load_module('family', ruledefs)
        self.engine.load(resource_filename(inputs.__name__, 'fathers.txt'))
        
    def test_df(self):
        df = facts_to_df(self.engine, 'IsFatherOf')
        names = 'Adam Seth Enos Kenan Mahalalel'.split()
        self.assertEqual(list(df['father']), names[:-1])
        self.assertEqual(list(df['son']), names[1:])



"""
This module implements aggregators, in which facts are collected during 
the rule engine execution and then aggregated in Python code afterwards. 
"""


from bunch import Bunch
from collections import defaultdict

from expr import FieldExpr
from fact import FactReference
from utils import RuleEngineError

class Aggregator(object):
    def __init__(self, engine, template):
        self.engine = engine
        self.template = template
        self.finalized = False
        
    def init(self):
        """ Runs before rule engine execution """
        pass
    
    def finalize(self, assert_):
        """ Runs after rule engine execution """
        self.finalized = True
    
    def process_one(self, **kwargs):
        """ Runs on each fact to be aggregated """
        raise NotImplementedError

def grouping_aggregator(keys, func):
    """ 
    'Groupby' aggregator. Facts are collected to lists grouped by
    (zero or more) keys and then each group is processed separately
    by the given 'func'
    """ 
    keys = tuple(_make_key(key) for key in keys)
    class GroupAggregator(Aggregator):
        def init(self): 
            self.data = defaultdict(lambda : [])
            
        def process_one(self, **kwargs):
            if self.finalized:
                raise RuleEngineError('Aggregation failed')
            key = tuple(key(**kwargs) for key in keys)
            value = kwargs.values()[0] if len(kwargs) == 1 else Bunch(kwargs)
            self.data[key].append(value)
            
        def finalize(self, assert_):
            super(GroupAggregator, self).finalize(assert_)
            for key, group in self.data.iteritems():
                func(group, assert_, *key)
        
    return GroupAggregator

def _make_key(key):
    if isinstance(key, FieldExpr):
        return lambda **kwargs : getattr(_make_key(key.container)(**kwargs), key.name)
    elif isinstance(key, FactReference):
        return lambda **kwargs : key._from_python_params(kwargs)
    else:
        raise TypeError('Invalid group key: {}'.format(type(key)))



import sys

class ProgressBar(object):
    def __init__(self, total, message=None, blocksize=100, width=20):
        self.total = total
        self.message = message or ''
        self.blocksize = blocksize
        self.width = width
        self.count = 0
        self.blocknum = -1
        
    def increment(self, value=1):
        prev_blocknum = self.blocknum
        self.count += value
        self.blocknum = self.count / self.blocksize
        finished = self.count == self.total
        if self.blocknum != prev_blocknum or finished:
            progress = self.count * self.width / self.total 
            output = '{} [{}{}] {}/{}\r'.format(self.message, '*' * progress,
                    '-'*(self.width - progress), self.count, self.total)
            self._emit(output)
        if finished:
            self._emit_finished() 
            
    def _emit(self, message): raise NotImplementedError
    def _emit_finished(self): raise NotImplementedError

class StderrProgressBar(ProgressBar):
    def _emit(self, message):
        print >>sys.stderr, message
        
    def _emit_finished(self):
        print >>sys.stderr
        print >>sys.stderr
        
class LogProgressBar(ProgressBar):
    def __init__(self, logger, *args, **kwargs):
        super(LogProgressBar, self).__init__(*args, **kwargs)
        self.logger = logger
        
    def _emit(self, message):
        self.logger.debug(message)
        
    def _emit_finished(self):
        pass



from collections import defaultdict
from functools import partial

from actions import Assert
from expr import FieldExpr, VariableExpr, IntegerField, normalize_expr
from fact import Fact, FactExpr, RULU_INTERNAL_PREFIX
from func import RuleFunc
from operators import Condition
from slots import make_slotted_type
from typedefs import TYPE_MAP, FactIndexType, Integer, Multifield
from utils import LispExpr, RuleEngineError, UniqueIdCounter, logger, wrap_clips_errors

class RulePremise(object):
    def __init__(self, container, var_name):
        self.container = container
        self.var_name = var_name
        self.negative = False
        self.field_values = set() # (field, value) pairs
        
    def add(self, field, value):
        assert field.container is self.container, (field.container, self.container)
        self.field_values.add((field, value))
        
    def set_negative(self):
        self.negative = True
        
    def build_lisp(self):
        field_values = sorted(self.field_values, key = lambda (field, value) : str(field))
        res = LispExpr(self.container._name, *(LispExpr(field.name, value) for field, value in field_values))
        if self.negative:
            res = LispExpr('not', res)
        return res
        
    def build_str(self, with_varname=True):
        res = str(self.build_lisp())
        if not res.startswith('('):
            res = '({})'.format(res)
        if with_varname and not self.negative:
            res = '{} <- {}'.format(self.var_name, res)
        return res
        
    def build_target_field(self):
        return LispExpr(self.target_field_name, self.target_field_var)

    def replace_values(self, mapping):
        self.field_values = {(field, mapping.get(value, value)) for field, value in self.field_values}


class Rule(object):
    def __init__(self):
        self.premises = defaultdict(RulePremise)
        self.name = None
        self.description = None
        self.target = None
        self.target_fields = None
        self.salience = None
        self.variable_map = {}
        self.variable_values = {}
        self.conditions = []
        self.actions = []
        self.python_actions = []
        self.aggregator_classes = []
        self.groupby = {}
        self.secondary_rules = []
        self.num_functions = 0
        self.comments = None
        
        self.premise_varnames = UniqueIdCounter('?fact')
        self.general_varnames = UniqueIdCounter('?var')
        
    def set_target(self, target=None):
        if self.target_fields is not None:
            raise RuleEngineError('Cannot set both target template and target fields')
        self.target = target
        self.target_fields = target._fields
        
    def set_name(self, name):
        if self.name not in (None, name):
            raise RuleEngineError('Tried to set target name to "{}", but it is already "{}"'.format(name, self.name))
        self.name = name
            
    def set_description(self, description):
        self.description = description
        
    def set_target_fields(self, **fields):
        if self.target is not None:
            raise RuleEngineError('Target already set')
        self.target_fields = {key: FieldExpr(_type=TYPE_MAP.get(_type, _type)) for key, _type in fields.iteritems()}
        self.target = make_slotted_type(Fact, self.name, **self.target_fields)
        
    def get_target_fields(self):
        return self.target_fields
    
    def get_target_name(self):
        if self.target is not None and self.target._name is not None:
            return self.target._name
        else:
            return self.name
        
    def set_salience(self, salience):
        self.salience = salience
        
    def add_variable(self, *exprs):
        free_value = None
        matching_var_names = set()
        for expr in sorted(exprs, key=str):
            if isinstance(expr, (FieldExpr, FactExpr)):
                try:
                    matching_var_names.add(self.variable_map[expr].var_name)
                except KeyError:
                    pass
            else:
                expr = normalize_expr(expr)
                if free_value is None:
                    free_value = expr
                elif free_value != expr:
                    raise ValueError('Conflicting unbound values: {}, {}'.format(free_value, expr))
        
        if not matching_var_names:
            var_name = self.general_varnames.next()
            if exprs[0].get_type() is Multifield:
                var_name = '$' + var_name
        else:
            var_names = sorted(matching_var_names)
            var_name = var_names[0]
            for prev_var_name in var_names[1:]:
                self._replace_variable(prev_var_name, var_name)

        for expr in exprs:
            if isinstance(expr, FactExpr):
                self.set_premise(expr.fact)
                if free_value is not None:
                    raise ValueError('Type mismatch: {}, {}'.format(expr, free_value))
            elif isinstance(expr, FieldExpr):
                self.set_premise(expr.container).add(expr, var_name)
                self.variable_map[expr] = VariableExpr(var_name, expr.get_type())
                if free_value is not None:
                    self.variable_values[var_name] = free_value.to_lisp()
        return var_name
    
    def add_condition(self, expr):
        condition = Condition(expr)
        for field in condition.all_fields:
            self.add_variable(field)
        self.conditions.append(condition)

    def add_action(self, action):
        for field in action.get_all_fields():
            self.add_variable(field)
        action.prepare_rule(self)
        self.actions.append(action)
                
    def add_python_action(self, function):
        if function.__name__ == '<lambda>':
            function.__name__ = 'lambda'
        self.python_actions.append(function)
        
    def add_aggregator_cls(self, aggregator_cls):
        self.aggregator_classes.append(aggregator_cls)
        
    def set_comments(self, comments):
        self.comments = comments

    def set_premise(self, container):
        premise = self.premises.get(container)
        if premise is None:
            var_name = self.premise_varnames.next()
            premise = RulePremise(container, var_name)
            self.premises[container] = premise
            self.variable_map[container._to_expr()] = VariableExpr(var_name, FactIndexType)
        return premise

    def set_negative_premise(self, container):
        self.set_premise(container).set_negative()
        
    def set_groupby(self, **kwargs):
        self.groupby.update(kwargs)
        
    def add_secondary_rule(self, rule):
        self.secondary_rules.append(rule)

    @wrap_clips_errors
    def build(self, engine):
        if self.target is not None:
            self._init_target(engine)
        self._init_aggregators(engine)
        if self.python_actions:
            target_name = self.target._name if self.target else None
            self._add_python_action(engine)
        if self.name is None:
            self.name = '_anonymous_rule'
        rule_num = sum(1 for rule_name in engine.get_rule_names() if rule_name.startswith(self.name+'@'))
        self.clips_name = '{}@{}'.format(self.name, rule_num+1)
        lhs = self._build_lhs()
        rhs = self._build_rhs()
        logger.getChild('rule').debug('Creating rule: %s\n<%s\n%s\n%s>\n%s',
                self.clips_name, '='*20, lhs, '='*20, rhs)
        self.clips_rule = engine.environment.BuildRule(self.clips_name, lhs, rhs, self.comments)
        for secondary_rule in self.secondary_rules:
            secondary_rule.build(engine)

    def _init_target(self, engine):
        target_name = self.get_target_name()
        if target_name is None:
            raise RuleEngineError('Cannot build rule, target name not set or multiple targets have the same name')
        if self.name is None:
            self.name = target_name
        if self.target._name is None:
            self.target._name = target_name
        if target_name not in engine.clips_types:
            self.target._build(engine)
        
    def _init_aggregators(self, engine):
        # TODO: Move this to separate module with secondary rule
        self.finalize_rules = []
        for aggregator_cls in self.aggregator_classes:
            aggregator = aggregator_cls(engine=engine, template=self.target)
            engine.preprocess_funcs.append(aggregator.init)
            self.add_python_action(lambda assert_, **kwargs : aggregator.process_one(**kwargs))
            finalize_name = '{}_{}_finalize'.format(RULU_INTERNAL_PREFIX, self.name)
            finalize_fact_cls = type(Fact)(finalize_name, (Fact, ), {'x': IntegerField()})
            finalize_fact_cls._build(engine)
            target = self.target
            self.target = finalize_fact_cls
            if self.salience is None: self.salience = -1000
            self.add_action(Assert(x=0))
            finalize_rule = Rule()
            finalize_rule.set_name(finalize_name)
            finalize_rule.set_premise(finalize_fact_cls)
            finalize_rule.set_salience(self.salience)
            finalize_rule.set_target(target)
            self.set_salience((self.salience)+1)
            finalize_rule.add_python_action(lambda assert_, **kwargs: aggregator.finalize(assert_))
            finalize_rule.build(engine)
            self.finalize_rules.append(finalize_rule)
            
    def _add_python_action(self, engine):
        func = RuleFunc(partial(self._action, engine), Integer, '_'.join(x.__name__ for x in self.python_actions))
        params = [premise.container for premise in self.premises.itervalues() if not premise.negative]
        self.actions.append(func(*params).replace_fields(self.variable_map))
        
    def _build_lhs(self):
        lhs = []
        if self.salience is not None:
            lhs.append(LispExpr('declare', LispExpr('salience', self.salience)))
        for premise in sorted(self.premises.itervalues(), key=lambda p:p.var_name):
            premise.replace_values(self.variable_values)
            lhs.append(premise.build_str())
        lhs.extend(condition.replace_fields(self.variable_map).to_lisp() for condition in self.conditions) 
        return '\n'.join(str(x) for x in lhs)
    
    def _build_rhs(self):
        actions_lisp = [action.replace_fields(self.variable_map).to_lisp() for action in self.actions]
        self._postprocess_actions(actions_lisp)
        return '\n'.join(map(str, actions_lisp))

    def _postprocess_actions(self, actions_lisp):
        # Placeholder for trace plugin
        pass

    def _make_field_map(self):
        return {}
    
    def _action(self, _engine, *facts):
        params = {}
        vars = [var for var, premise in self.premises.iteritems() if not premise.negative]
        for var, fact in zip(vars, facts):
            var._update_python_param(params, fact)
        if self.target is not None:
            params['assert_'] = lambda **kwargs : _engine.assert_(self.target, **kwargs)
        self._run_python_actions(**params)

    def _run_python_actions(self, **params):
        for func in self.python_actions:
            try:
                func(**params)
            except:
                logger.error('Exception while running {} in rule {}.'.format(func, self.name))
                raise



from expr import FieldExpr
from utils import logger, wrap_clips_errors, RuleEngineError
from typedefs import Multifield

ANONYMOUS = 'Anonymous'

class SlotsMeta(type):
    def __new__(meta, name, bases, attrs): #@NoSelf
        attrs['_fields'] = {k: v for k, v in attrs.iteritems() if isinstance(v, FieldExpr)}
        for base in bases:
            if type(base) is meta:
                attrs['_fields'].update(base._fields)
        attrs['_name'] = None if name == ANONYMOUS else name
        attrs['__slots__'] = ('_clips_obj', '_data')
        cls = super(SlotsMeta, meta).__new__(meta, name, bases, attrs)
        cls._instrument()
        return cls
    
    def __repr__(self):
        return '<{}>'.format(self._name or ANONYMOUS)
        
class HasSlots(object):
    __metaclass__ = SlotsMeta
    _all_subclasses = []
    
    @classmethod
    def _instrument(cls):
        for key, value in cls._fields.iteritems():
            value._init(container=cls, field_name=key)
        cls._slots = ' '.join(cls._make_slot(key, value._type)
                              for key, value in cls._fields.iteritems())
        cls._clips_type = None # To be filled in _build()
        cls._all_subclasses.append(cls)
        cls._ordered_fields = sorted(cls._fields, key=lambda field : cls._fields[field].order)
        cls._field_order = {name: n for n, name in enumerate(cls._ordered_fields)}
        
    @classmethod
    def _make_slot(cls, name, type):
        if issubclass(type, Multifield):
            return '(multislot {})'.format(name)
        else:
            return '(slot {} (type {}))'.format(name, type.CLIPS_TYPE)

    @classmethod
    def _build(cls, engine):
        if not cls._fields:
            raise RuleEngineError('{} has no fields'.format(cls._name))
        logger.getChild('slots').debug('Building: %s, slots: %s', cls._name, cls._slots)
        cls._environment = engine.environment
        engine.register_clips_type(cls)
        
    @classmethod        
    def _indexed_reference(self, index):
        raise NotImplementedError
    
    @wrap_clips_errors
    def __init__(self, **kwargs):
        if self._clips_type is None:
            raise RuleEngineError('Class {} is not bound to a rule engine and cannot be instantiated'.format(type(self)))
        existing_clips_obj = kwargs.get('_clips_obj')
        if existing_clips_obj is None:
            self._clips_obj = self._create_clips_obj()
            self._clips_obj.AssignSlotDefaults()
            for key, value in kwargs.iteritems():
                field = self._fields.get(key)
                if field is None:
                    raise RuleEngineError('Field {} does not exist in {}'.format(key, self._name))
                if not field.get_type()._isinstance(value):
                    raise RuleEngineError('In {}: expected {} should be of type {}, got {}.'.
                                          format(self._name, key, field.get_type(), value))
                if isinstance(value, HasSlots):
                    value = value._clips_obj
                self._clips_obj.Slots[key] = field._type._to_clips_value(value)
        else:
            self._clips_obj = self._copy_clips_obj(existing_clips_obj)
        self._data = tuple(self._fields[key]._type._from_clips_value(self._clips_obj.Slots[key]) 
                           for key in self._ordered_fields)

    def _create_clips_obj(self):
        raise NotImplementedError
    
    def _copy_clips_obj(self, existing_clips_obj):
        raise NotImplementedError
    
    def __getattribute__(self, key):
        if key != '_fields' and key in self._fields:
            return self._data[self._field_order[key]]
        else:
            return object.__getattribute__(self, key)

    def _as_dict(self):
        return {key : getattr(self, key) for key in self._fields}
    
    def __repr__(self):
        return '<{}: {}>'.format(self._name, self._as_dict())
    
def make_slotted_type(cls, name=None, **fields):
    return SlotsMeta(name or ANONYMOUS, (cls, ), dict(fields))
    



import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand
from os import path


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)
    

setup(
    name='rulu',
    version='0.1.3',
    url='http://github.com/ubarkai/rulu',
    license='LGPL',
    author='Uri Barkai',
    author_email='ubarkai@gmail.com',
    description='Python interface for building rule-based expert systems over PyCLIPS',
    install_requires = ['bunch', 'pyclips'],
    tests_require=['pytest'],
    extras_require={'testing': ['pytest']},
    cmdclass={'test': PyTest},
    test_suite='test.test_ruleengine',
    packages=['rulu']
)



import clips
from actions import Update
from slots import HasSlots, SlotsMeta
from expr import BaseExpr, ConvertibleToExpr
from typedefs import FactIndexType
from utils import Printable, wrap_clips_errors

RULU_INTERNAL_PREFIX = '_rulu_internal'

class FactReference(ConvertibleToExpr):
    def _update_python_param(self, params, value):
        raise NotImplementedError
    
    def _from_python_params(self, params):
        raise NotImplementedError

class FactMeta(SlotsMeta, FactReference):
    def __getitem__(self, key):
        return self._indexed_reference(key)
    
indexed_references = {}

class Fact(HasSlots):
    __metaclass__ = FactMeta
    CLIPS_INSTANCE_TYPE = clips.Fact
            
    @classmethod
    @wrap_clips_errors
    def _build(cls, engine):
        if cls is not Fact: 
            super(Fact, cls)._build(engine)
            cls._clips_type = engine.environment.BuildTemplate(cls._name, cls._slots)
        
    @classmethod
    def _to_expr(cls):
        if '_fact_expr' not in cls.__dict__:
            cls._fact_expr = FactExpr(cls)
        return cls._fact_expr

    def _create_clips_obj(self):
        return self._clips_type.BuildFact()

    def _copy_clips_obj(self, obj):
        return obj if hasattr(obj, '_Fact__env') else self._environment.Fact(obj)
    
    def _clips_index(self):
        return self._clips_obj.Index
    
    def _delete(self):
        self._clips_obj.Retract()
        
    def _update(self, **kwargs):
        cmd = Update(self._clips_obj.Index, **kwargs).to_lisp()
        self._environment.SendCommand(str(cmd))
        
    @classmethod
    def _update_python_param(cls, params, value):
        params[cls._name] = value
        
    @classmethod
    def _from_python_params(cls, params):
        return params[cls._name]

    @classmethod
    def _indexed_reference(cls, index):
        return indexed_references.setdefault((cls, index), IndexedFactReference(cls, index))

class IndexedFactReference(Printable, FactReference):
    def __init__(self, template_cls, key):
        self._template_cls = template_cls
        self._key = key
        self._fields = {}
        self._fact_expr = FactExpr(self)
        for key, field in template_cls._fields.iteritems():
            new_field = type(field)(_type=field.get_type())
            new_field._init(self, key)
            self._fields[key] = new_field
            setattr(self, key, new_field)

    def _to_expr(self):
        return self._fact_expr

    def __str__(self):
        return '{}[{}]'.format(self._template_cls._name, repr(self._key))
    
    def _update_python_param(self, params, value):
        params.setdefault(self._template_cls._name, {})[self._key] = value
    
    def _from_python_params(self, params):
        return params[self._name][self._key]

    @property
    def _name(self):
        return self._template_cls._name
    
class FactExpr(BaseExpr):
    def __init__(self, fact):
        super(FactExpr, self).__init__(all_fields=[self])
        self.fact = fact
        
    def __str__(self):
        return str(self.fact)
    
    def get_type(self):
        return FactIndexType

    def to_lisp(self):
        return str(self)



from expr import BaseExpr, VariableExpr, normalize_expr
from typedefs import Boolean, Integer, Number
from utils import LispExpr, RuleEngineError

class BaseBinaryOperator(BaseExpr):
    def __init__(self, op, *args):
        super(BaseBinaryOperator, self).__init__(all_fields=set().union(*(arg.all_fields for arg in args)))
        self.op = op
        self.args = args
        
    def to_lisp(self):
        return LispExpr(self.op, *(arg.to_lisp() for arg in self.args))
    
    def replace_fields(self, field_map):
        return type(self)(self.op, *(arg.replace_fields(field_map) for arg in self.args))
        
    def __str__(self):
        return '({} {} {})'.format(self.lhs, self.op, self.rhs)
    
class ArithmeticBinaryOperator(BaseBinaryOperator, BaseExpr):
    def get_type(self):
        lhs_type, rhs_type = [arg.get_type() for arg in self.args]
        if lhs_type in (Integer, Number) and rhs_type in (Integer, Number):
            return Integer if lhs_type is Integer and rhs_type is Integer else Number
        else:
            raise RuleEngineError('Illegal operator "{}" on {} and {}'.format(
                    self.op.__name__, self.lhs.get_type(), self.rhs.get_type()))
    
class BooleanBinaryOperator(BaseBinaryOperator, BaseExpr):
    def get_type(self):
        return Boolean
    
class Condition(BaseExpr):
    def __init__(self, expr):
        expr = normalize_expr(expr)
        super(Condition, self).__init__(all_fields=expr.all_fields)
        if expr.get_type() is not Boolean:
            raise RuleEngineError('Condition must be of boolean type. Got "{}", which is {}'.format(expr, expr.get_type()))
        self.expr = expr
        
    def get_type(self):
        return Boolean

    def to_lisp(self):
        expr = self.expr
        if isinstance(self.expr, VariableExpr):
            expr = not_(not_(expr))
        return LispExpr('test', expr.to_lisp())
    
    def replace_fields(self, field_map):
        return type(self)(self.expr.replace_fields(field_map))
    
    def __str__(self):
        return 'Condition <{}>'.format(self.expr)
    
class UnaryOperator(BaseExpr):
    def __init__(self, expr, op):
        super(UnaryOperator, self).__init__(all_fields=expr.all_fields)
        self.expr = expr
        self.op = op
        
    def get_type(self):
        return self.expr.get_type()
    
    def to_lisp(self):
        return LispExpr(self.op, self.expr.to_lisp())
    
    def replace_fields(self, field_map):
        return type(self)(expr=self.expr.replace_fields(field_map), 
                          op=self.op)
        
    def __str__(self):
        return '({} {})'.format(self.op, self.expr)

or_ = lambda *args: BooleanBinaryOperator('or', *(normalize_expr(arg) for arg in args))
and_ = lambda *args: BooleanBinaryOperator('and', *(normalize_expr(arg) for arg in args))
not_ = lambda expr: UnaryOperator(normalize_expr(expr), op='not')
